unconfigured <- function() {
  fork  <- get("use.FORK",  envir = cluster)
  mpi   <- get("use.MPI",   envir = cluster)
  psock <- get("use.PSOCK", envir = cluster)
  if (!fork & !mpi & !psock) {
    stop("You need to configure ctools if you're going to use it.")
  }
}

getCores <- function(fname = NULL, cores = NULL) {

  mpiuniverse <- Sys.getenv("OMPI_UNIVERSE_SIZE")
  if ((mpiuniverse != "") & is.null(cores)) {
    cores <- as.numeric(mpiuniverse)
  }

  if (is.null(fname)) {
    if (!is.numeric(cores)) {
      cores <- parallel::detectCores()
    }
    hosts <- rep("localhost", cores)
    out <- list(cores = cores, hosts = hosts)
  } else if (!file.exists(fname)) {
    if (!is.numeric(cores)) {
      cores <- parallel::detectCores()
    }
    hosts <- rep("localhost", cores)
    out <- list(cores = cores, hosts = hosts)
  } else {
    mpi_slots <- utils::read.table(fname)
    core_num <- as.character(mpi_slots[,2])
    hostnames <- as.character(mpi_slots[,1])
    cores_count <- 0
    hosts <- c()
    for(i in 1:length(core_num)) {
      core <- core_num[i]
      host <- hostnames[i]
      ss <- as.numeric(strsplit(core, "=")[[1]][2])
      cores_count <-  cores_count + ss
      hosts <- c(hosts, rep(host, ss))
    }
    if (is.numeric(cores)) {
      hosts <- hosts[1:cores]
    } else {
      cores <- cores_count
    }
  }
  out <- list(cores = cores, hosts = hosts)
  return(out)
}

#' Configure options for cores directly
#'
#' ctools sets the default cores number to the maximum possible,
#' this function allows the user to change the number of cores used
#' in local-only parallel, i.e. PSOCKS clusters and FORK parallel.
#' NOTE: Changing cores for PSOCKS, used exclusively on Windows machines,
#' restarts the cluster, thus all exported objects are lost. Thus, it's
#' recommended that this function be used shortly after loading the ctools
#' package. Unix-like machines, e.g. Linux or OSX, use forking, so the core
#' number can be changed at any time
#' @param type the type of cluster you'd like to use. This can be configured with optins(ctools_type = "<type>").
#' @param hostfile if using PSOCK, an MPI style hostfile specifiying the hostnames and slot numbers for the cluster.
#' @param cores the number of cores to assign, defaults to maximum number.
#' @export c.config
#' @examples
#' c.config(cores = 2)		# Set cores number to 2
c.config <- function(type = c("MPI", "FORK", "PSOCK"), hostfile = NULL, cores = NULL) {

  type     <- getOption("ctools_type", default = type)
  hostfile <- getOption("ctools_hostfile", default = hostfile)
  cores    <- getOption("ctools_cores", default = cores)

  conf <- getCores(fname = hostfile, cores = cores)

  assign("use.MPI", F, envir = cluster)
  assign("use.FORK", F, envir = cluster)
  assign("use.cluster", F, envir = cluster)

  for (i in type) {
    if ((i == "MPI") & (get("mpi.check", envir = cluster))) {
			assign("use.MPI", T, envir = cluster)
			assign("use.cluster", T, envir = cluster)
			use.MPI(conf$cores)
			break
		} else if ((i == "FORK") & get("is.unix", envir = cluster)) {
			assign("use.FORK", T, envir = cluster)
			assign("use.cluster", F, envir = cluster)
			use.FORK(cores = conf$cores)
			break
		} else if (i == "PSOCK") {
			assign("use.PSOCK", T, envir = cluster)
			assign("use.cluster", T, envir = cluster)
			use.PSOCKS(cores = conf$cores, hosts = conf$hosts)
			break
		} 
  }
}

