#' Portable package loader
#'
#' If utilizing FORK style parallel, load all packages in the
#' package list into the main R process. If utilizing a cluster, additionally
#' load the packages into all cluster nodes.
#' @param pkgs character vector of package names to load
#' @param lib.loc a character vector describing the loaction of R library trees to search through,
#' or 'NULL'. The default value of 'NULL' corresponds to all libraries currently know to '.libPaths()'/
#' Non-existant library trees are silently ignored.
#' @param quietly a logical. If 'TRUE', no message confirming package attaching is printed.
#' @export c.library
#' @examples
#' c.libary("ggplot2","reshape2")
c.library <- function(pkgs, lib.loc = NULL, quietly = TRUE) {

	if (get("use.cluster", envir = cluster)) {
		cl.call <- get("cluster.call", envir = cluster)
		cl <- get("cluster.object", envir = cluster)
		capture <- lapply(pkgs, require, character.only = TRUE, quietly = quietly, lib.loc = lib.loc)
		capture <- cl.call(cl, function(pkgs, quietly, lib.loc) {
									lapply(pkgs ,require, character.only = TRUE, quietly = quietly, lib.loc = lib.loc)
								}, pkgs, quietly = quietly, lib.loc = lib.loc)
	} else {
		capture <- lapply(pkgs, require, character.only = TRUE, quietly = quietly, lib.loc = lib.loc)
	}

}

#' Return the number of cores configured for use
#'
#' This function returns the number of cores configured for use with ctools
#' @export c.cores
#' @param local return only the number of cores on the local machine
#' @examples
#' cores <- c.cores() # Return the number of cores in use
c.cores <- function(local = F) {
  if (local) {
    parallel::detectCores()
  } else {
    get("cores", envir = cluster)
  }
}

#' Return the maximum number of cores possible to use
#'
#' This function returns the maximum number of cores possible to use with ctools
#' @export c.maxcores
#' @examples
#' maxcores <- c.maxcores()		# Return the maximum number of cores available
c.maxcores <- function() {
	get("maxcores", envir = cluster)
}

#' Return the number of cores configured for use
#'
#' Set seet for random number generators.
#' @export c.set.seed
#' @examples
#' c.set.seed(42) # set the seed on all workers
c.set.seed <- function(seed) {

	if (get("use.mpi", envir = cluster)) {
		cl <- get("cluster.object", envir = cluster)
		cl.seed <- get("cluster.seeder", envir = cluster)
		cl.seed(cl, seed)
	} else if (get("use.cluster", envir = cluster)) {
		cl.call <- get("cluster.call", envir = cluster)
		cl <- get("cluster.object", envir = cluster)
		cl.call(cl, set.seed, seed=seed )
	}
	set.seed(seed)
}
